# Get A Clue (C) 2010 V. Harishankar
# Crossword puzzle maker program
# Licensed under the GNU GPL v3

# Class for the puzzle data representation

class GridItem:
	# initialize the item
	def __init__ (self, item_char='.', across_start = False, down_start = False,
					occupied_across = False, occupied_down = False, num = 0,
					clue_across = None, clue_down = None, revealed = False):
		# character in the cell
		self.char = item_char
		# is the cell the start of an across word?
		self.across_start = across_start
		# is the cell the start of a down word?
		self.down_start = down_start
		# is the cell occupied by a letter in an across word?
		self.occupied_across = occupied_across
		# is the cell occupied by a letter in a down word?
		self.occupied_down = occupied_down
		# numbering of the cell if it is the start of a word
		self.numbered = num
		# clue across if the cell is the start of an across word
		self.clue_across = clue_across
		# clue down if the cell is the start of a down word
		self.clue_down = clue_down
		# is the letter revealed or hidden?
		self.revealed = revealed

# exception for too long words
class TooLongWordException (Exception):
	def __init__ (self, word, length):
		self.word = word
		self.length = length

# exception for intersecting words
class IntersectWordException (Exception):
	def __init__ (self, word, length):
		self.word = word
		self.length = length

# exception when grid is sought to be changed when frozen
class FrozenGridException (Exception):
	def __init__ (self):
		self.msg = "Grid is frozen and cannot be edited"

# exception when no word is found at a position
class NoWordException (Exception):
	def __init__ (self, row, col):
		self.pos = (row, col)

# exception when no words are present in the grid
class NoWordsException (Exception):
	def __init__ (self):
		self.msg = "No words in grid"

class CrosswordPuzzle:
	def __init__ (self, rows, cols):
		# define number of rows and columns
		self.rows = rows
		self.cols = cols

		# initialize the list to hold the grid
		self.data = []

		# initial state of the grid is unfrozen
		self.frozen_grid = False

		# create the grid data
		for i in range (rows):
			self.data.append ([])
			for j in range (cols):
				self.data[i].append (GridItem ())

	# get all the clues for across
	def get_clues_across (self):
		clues = []
		# traverse the grid
		for row in range (self.rows):
			for col in range (self.cols):
				if (self.data[row][col].occupied_across is True and
				self.data[row][col].across_start is True):
					word_across = self.get_word_across (row, col)
					clues.append ((word_across, self.data[row][col].clue_across))
		# if no across words are found at all
		if not clues:
			raise NoWordsException

		return clues

	# get all the clues for down
	def get_clues_down (self):
		clues = []
		# traverse the grid
		for row in range (self.rows):
			for col in range (self.cols):
				if (self.data[row][col].occupied_down is True and
				self.data[row][col].down_start is True):
					word_down = self.get_word_down (row, col)
					clues.append ((word_down, self.data[row][col].clue_down))
		# if no down words are found at all
		if not clues:
			raise NoWordsException

		return clues

	# getting a down word at a position
	def get_word_down (self, row, col):
		# if index is out of bounds
		if row >= self.rows or col >= self.cols:
			raise NoWordException (row, col)

		# if there is no occupied down letter at that position
		if self.data[row][col].occupied_down is False:
			raise NoWordException (row, col)

		# now traverse the grid to find the beginning of the word
		i = row
		while i >= 0:
			# if it is occupied down and is the beginning of the word
			if (self.data[i][col].occupied_down is True and
				self.data[i][col].down_start is True):
				start_row = i
				break
			i -= 1

		i = start_row
		word_chars = []
		# now seek the end of the word
		while i < self.rows:
			if self.data[i][col].occupied_down is True:
				word_chars.append (self.data[i][col].char)
			else:
				break
			i += 1

		word = "".join (word_chars)

		# return the word, starting row, column and length as a tuple
		return (word, start_row, col, len(word))

	# getting an across word at a position
	def get_word_across (self, row, col):
		# if index is out of bounds
		if row >= self.rows or col >= self.cols:
			raise NoWordException (row, col)

		# if there is no occupied across letter at that position
		if self.data[row][col].occupied_across is False:
			raise NoWordException (row, col)

		# now traverse the grid to look for the beginning of the word
		i = col
		while i >= 0:
			# if it is occupied across and is the beginning of the word
			if (self.data[row][i].occupied_across is True and
				self.data[row][i].across_start is True):
				start_col = i
				break
			i -= 1

		i = start_col
		word_chars = []
		# now seek the end of the word
		while i < self.cols:
			if self.data[row][i].occupied_across is True:
				word_chars.append (self.data[row][i].char)
			else:
				break
			i += 1

		word = "".join (word_chars)

		# return the word, starting column, row and length as a tuple
		return (word, row, start_col, len(word))

	# setting a down word
	def set_word_down (self, row, col, word):
		# if the grid is frozen the abort
		if self.frozen_grid is True:
			raise FrozenGridException

		# if the word length greater than totalrows - startrow
		if len(word) > self.rows - row:
			raise TooLongWordException (word, len(word))

		# is the word intersecting any other word?
		for i in range (len(word)):
			# on the same column
			if self.data[row+i][col].occupied_down is True:
				raise IntersectWordException (word, len(word))
			# on the previous column except first column
			if col > 0:
				# except the first and last col
				if i > 0 and i < len(word) - 1:
					if self.data[row+i][col-1].occupied_down is True:
						raise IntersectWordException (word, len(word))
			# on the next column except last column
			if col < len(word) - 1:
				# except the first and last row check if there is any
				# down word in previous column
				if i > 0 and i < len(word) - 1:
					if self.data[row+i][col+1].occupied_down is True:
						raise IntersectWordException (word, len(word))
				# check if there is any across word starting in the
				# next column
				if self.data[row+i][col+1].across_start is True:
					raise IntersectWordException (word, len(word))

		# also check the character before and after
		if (row > 0 and self.data[row-1][col].occupied_down is True
			and self.data[row-1][col].occupied_across is True):
			raise IntersectWordException (word, len(word))
		if (row + len(word) < self.rows and
			self.data[row+len(word)][col].occupied_across is True and
			self.data[row+len(word)][col].occupied_down is True):
			raise IntersectWordException (word, len(word))

		# set the down start to true
		self.data[row][col].down_start = True
		# set the word
		for i in range (len(word)):
			self.data[row+i][col].occupied_down = True
			self.data[row+i][col].char = word[i].upper ()


	# setting an across word
	def set_word_across (self, row, col, word):
		# if the grid is frozen the abort
		if self.frozen_grid is True:
			raise FrozenGridException

		# is the word length greater than totalcols - startcol?
		if len(word) > self.cols - col:
			raise TooLongWordException (word, len(word))

		# is the word intersecting any other word?
		for i in range (len(word)):
			# on the same row
			if self.data[row][col+i].occupied_across is True:
				raise IntersectWordException (word, len(word))
			# on a previous row except first row
			if row > 0:
				# if not the first or last col
				if i > 0 and i < len(word) - 1:
					if self.data[row-1][col+i].occupied_across is True:
						raise IntersectWordException (word, len(word))
			# on a next row
			if (row < (self.rows - 1)):
				# except the first and last letter check if there is
				# any across intersection
				if i > 0 and i < len (word) - 1:
					if self.data[row+1][col+i].occupied_across is True:
						raise IntersectWordException (word, len(word))
				# if a down word is starting at any column below the
				# word
				if self.data[row+1][col+i].down_start is True:
					raise IntersectWordException (word, len(word))

		# also check the character beyond and before and after
		if (col > 0 and (self.data[row][col-1].occupied_across is True or
					self.data[row][col-1].occupied_down is True)):
			raise IntersectWordException (word, len(word))
		if (col + len(word) < self.cols and
				(self.data[row][col+len(word)].occupied_across is True or
				self.data[row][col+len(word)].occupied_down is True)):
			raise IntersectWordException (word, len(word))

		# set across start to true
		self.data[row][col].across_start = True

		# set the word
		for i in range (len(word)):
			self.data[row][col+i].char = word[i].upper ()
			self.data[row][col+i].occupied_across = True

	# freeze the grid numbers etc.
	def freeze_grid (self):
		# numbering
		numbering = 1
		# run through the grid
		for row in range (self.rows):
			for col in range (self.cols):
				# if grid is blank set the character to #
				if (self.data[row][col].occupied_across is False
					and self.data[row][col].occupied_down is False):
					self.data[row][col].char = "#"
				elif (self.data[row][col].across_start is True or
					self.data[row][col].down_start is True):
					self.data[row][col].numbered = numbering
					numbering += 1

		self.frozen_grid = True

	# unfreeze the grid numbers etc.
	def unfreeze_grid (self):
		# run through the grid
		for row in range (self.rows):
			for col in range (self.cols):
				self.data[row][col].numbered = 0
				if (self.data[row][col].occupied_across is False and
					self.data[row][col].occupied_down is False):
					self.data[row][col].char = '.'

		self.frozen_grid = False

